pub enum MinimizationCutoff {
None,
Iteration(usize),
Decrease(f64),
BreakAfterFirstImprovement,
}Expand description
Denotes when to stop searching for better vtree fragment states when dynamically minimizing SDDs.
Variants
None
Does not stop and explores all 12 fragment states.
Iteration(usize)
Stops after i-th iteration.
Decrease(f64)
Stops after decrease in size of a referential SDD reaches given threshold:
size(after)/size(before) <= threshold. Therefore, if size of SDD drops
from 150 to 100 and the threshold is set to 0.8, the search terminates
since 100/150 = 0.66 <= 0.8.
BreakAfterFirstImprovement
Stop minimizing right after encountering first fragment state that improves the size of the referentail SDD.